# -*- coding: utf-8 -*-

#    Copyright (C) 2019  Marcus Rickert
#
#    See https://github.com/marcus67/python_base_app
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import datetime
import time
import inspect
import io
import json
import os
import platform
import pwd
import re
import stat
import sys
import threading
import traceback
import urllib.parse
from os.path import dirname

from python_base_app import configuration
from python_base_app import exceptions
from python_base_app import log_handling

FORMAT_TIME_WITH_SECONDS = "%H:%M:%S"
FORMAT_TIME = "%H:%M"
FORMAT_DATE = "%Y-%m-%d (%a)"
FORMAT_SHORT_DATE = "%A"
FORMAT_SIMPLE_DATE = "%Y-%m-%d"
FORMAT_DATETIME = "%Y-%m-%d %H:%M:%S"
FORMAT_DURATION_WITH_SECONDS = "%dh%02dm%02ds"
FORMAT_DURATION = "%dh%02dm"

FORMAT_JSON_DATETIME = "%Y-%m-%d %H:%M:%S"

REGEX_DURATION = re.compile("^ *(([0-9]+) *[h|H])? *(([0-9]+) *[m|M])? *(([0-9]+) *[s|S])? *$")
REGEX_TIME = re.compile("^ *([0-9]+)( *: *([0-9]+))?( *: *([0-9]+))? *$")

EMPTY_DURATION = "-"
EMPTY_TIME = "-"

PLATFORM_NAME_WINDOWS = 'windows'
PLATFORM_NAME_MAC_OS = "darwin"

PASSWORD_PATTERNS = ("PASSW", "KENNW", "ACCESS")
PROTECTED_PASSWORD_VALUE = "[HIDDEN]"


def int_to_string(p_integer):
    if p_integer is None:
        return "[None]"

    else:
        return str(p_integer)


def is_windows():
    return platform.system().lower() == PLATFORM_NAME_WINDOWS


def is_mac_os():
    return platform.system().lower() == PLATFORM_NAME_MAC_OS


def get_current_time():
    return datetime.datetime.now()


def get_date_as_string(p_date, p_short=False):
    if p_date is not None:
        if p_short:
            return p_date.strftime(FORMAT_SHORT_DATE)

        else:
            return p_date.strftime(FORMAT_DATE)

    else:
        return "[None]"


def get_simple_date_as_string(p_date):
    if p_date is not None:
        return p_date.strftime(FORMAT_SIMPLE_DATE)

    else:
        return "[None]"


def get_timestamp_as_string(p_timestamp):
    if p_timestamp is not None:
        return p_timestamp.strftime(FORMAT_DATETIME)

    else:
        return "[None]"


def get_time_as_string(p_timestamp, p_include_seconds=True):
    if p_timestamp is not None:
        if p_include_seconds:
            return p_timestamp.strftime(FORMAT_TIME_WITH_SECONDS)

        else:
            return p_timestamp.strftime(FORMAT_TIME)

    else:
        return "[None]"


def get_duration_as_string(p_seconds, p_include_seconds=True):
    if p_seconds is None:
        return "-"

    else:
        hours = int(p_seconds / 3600)
        minutes = int((p_seconds - hours * 3600) / 60)

        if p_include_seconds:
            seconds = int(p_seconds - hours * 3600 - minutes * 60)
            return FORMAT_DURATION_WITH_SECONDS % (hours, minutes, seconds)

        else:
            return FORMAT_DURATION % (hours, minutes)


def is_protected_name(p_name):
    for pattern in PASSWORD_PATTERNS:
        if pattern in p_name.upper():
            return True

    return False


def protect_password_value(p_name, p_value):
    if is_protected_name(p_name=p_name):
        return PROTECTED_PASSWORD_VALUE

    else:
        return p_value


def get_safe_attribute_name(p_string):
    return p_string.replace(":", "_").replace("-", "_")


def convert_query_result_to_json(p_result, p_columns):
    a = []
    for entry in p_result:
        d = {}
        i = 0
        for col in p_columns:
            raw_value = getattr(entry, col)
            if type(raw_value) is datetime.datetime:
                value = raw_value.strftime('%Y-%m-%d %H:%M:%S')
            else:
                value = raw_value
            d[col] = value
            i = i + 1
        a.append(d)

    return json.dumps(a, separators=(',', ':'))


def test_mode(p_filename, p_app_owner, p_executable=False, p_writable=False, p_is_directory=False, p_other_access=True):
    logger = log_handling.get_logger()
    p_filename = os.path.abspath(p_filename)

    if p_is_directory:
        if not os.path.isdir(p_filename):
            raise exceptions.InstallationException("Directory '%s' not found!" % p_filename)

    else:
        if not os.path.isfile(p_filename):
            raise exceptions.InstallationException("File '%s' not found!" % p_filename)

    info = os.stat(p_filename)

    if p_app_owner is not None:
        owner_name = pwd.getpwuid(info.st_uid).pw_name

        if owner_name != p_app_owner:
            raise exceptions.InstallationException(
                "File/directory '%s' must be owned by '%s'!" % (p_filename, p_app_owner))

        logger.info("File/directory '%s' is owned by '%s' -> OK" % (p_filename, p_app_owner))

    if info.st_mode & stat.S_IRUSR == 0:
        raise exceptions.InstallationException("File/directory '%s' must be readable by owner!" % p_filename)

    logger.info("File/directory '%s' must be readable by owner -> OK" % p_filename)

    if p_executable:
        if info.st_mode & stat.S_IXUSR == 0:
            raise exceptions.InstallationException("File/directory '%s' must be executable by owner!" % p_filename)
        logger.info("File/directory '%s' must be executable by owner -> OK" % p_filename)

    if p_writable:
        if info.st_mode & stat.S_IWUSR == 0:
            raise exceptions.InstallationException("File/directory '%s' must be writable by owner!" % p_filename)
        logger.info("File/directory '%s' must be writable by owner -> OK" % p_filename)

    if not p_other_access:
        if info.st_mode & (stat.S_IRGRP | stat.S_IROTH | stat.S_IWGRP | stat.S_IWOTH | stat.S_IXGRP | stat.S_IXOTH) > 0:
            raise exceptions.InstallationException(
                "File/directory '%s' must not be accessible by group or others!" % p_filename)
        logger.info("File/directory '%s' must not be accessible by group or others!" % p_filename)


class ObjectEncoder(json.JSONEncoder):

    def __init__(self, *args, **kwargs):

        super(ObjectEncoder, self).__init__(*args, **kwargs)

    def default(self, obj):

        if type(obj) is datetime.datetime:
            return self.default(obj.strftime(FORMAT_JSON_DATETIME))

        elif hasattr(obj, "to_json"):
            return self.default(obj.to_json())

        elif hasattr(obj, "__dict__"):
            d = dict(
                (key, value)
                for key, value in inspect.getmembers(obj)
                if not key.startswith("__")
                and not inspect.isabstract(value)
                and not inspect.isbuiltin(value)
                and not inspect.isfunction(value)
                and not inspect.isgenerator(value)
                and not inspect.isgeneratorfunction(value)
                and not inspect.ismethod(value)
                and not inspect.ismethoddescriptor(value)
                and not inspect.isroutine(value)
                and not isinstance(value, property)
            )
            return self.default(d)

        return obj


def anonymize_args(p_args):
    new_args = []
    hide = False

    for arg in p_args:
        if is_protected_name(arg):
            new_args.append(arg)
            hide = True
        elif hide:
            new_args.append(PROTECTED_PASSWORD_VALUE)
            hide = False
        else:
            new_args.append(arg)

    return new_args


def anonymize_url(p_url):
    components = urllib.parse.urlparse(p_url)

    if components.password is not None:
        fmt = "{username}:{protected_value}@{hostname}"
        net_location = fmt.format(username=components.username, protected_value=PROTECTED_PASSWORD_VALUE,
                                  hostname=components.hostname)
    else:
        net_location = components.netloc

    return urllib.parse.urlunsplit(
        (
            components.scheme,
            net_location,
            components.path,
            components.params,
            components.query
        ))


def get_package_directory(p_package):
    return dirname(p_package.__file__)


def check_config_value(p_config, p_config_attribute_name):
    if getattr(p_config, p_config_attribute_name) is None:
        raise configuration.ConfigurationException(
            "Setting [%s]%s is missing!" % (p_config.section_name, p_config_attribute_name))


def log_stack_trace(p_logger=None):
    (_type, _value, tb) = sys.exc_info()
    string_buffer = io.StringIO()
    traceback.print_tb(tb=tb, file=string_buffer)

    fmt = "Stack trace = %s" % str(string_buffer.getvalue())

    if p_logger is not None:
        p_logger.error(fmt)
    else:
        sys.stderr.write(fmt)


def handle_fatal_exception(p_exception, p_logger=None):
    if p_logger is not None:
        p_logger.fatal(str(p_exception))

    else:
        sys.stderr.write(str(p_exception))


def objectify_dict(p_dict, p_class, p_attribute_classes={}):
    instance = p_class()

    for attr_name, attr_value in p_dict.items():
        if attr_value is not None and attr_name in p_attribute_classes:
            attr_class = p_attribute_classes[attr_name]

            if attr_class == datetime.datetime:
                attr_value = datetime.datetime.strptime(attr_value, FORMAT_JSON_DATETIME)

        setattr(instance, attr_name, attr_value)

    return instance


def start_simple_thread(method, *args, **kwargs):
    new_thread = threading.Thread(target=method, args=args, kwargs=kwargs)
    new_thread.start()

    return new_thread


def get_string_as_duration(p_string):
    if p_string is None:
        return None

    if p_string.strip() in (EMPTY_DURATION, ''):
        return None

    match = REGEX_DURATION.match(p_string)

    if match is None:
        fmt = "Use Hh MMm SSs"
        raise configuration.ConfigurationException(fmt.format(string=p_string))

    result = 0

    if match.group(2) is not None:
        result = result + 3600 * int(match.group(2))

    if match.group(4) is not None:
        result = result + 60 * int(match.group(4))

    if match.group(6) is not None:
        result = result + int(match.group(6))

    return result


def get_string_as_time(p_string):
    if p_string is None:
        return None

    if p_string.strip() in (EMPTY_TIME, ''):
        return None

    match = REGEX_TIME.match(p_string)

    if match is None:
        fmt = "Use HH[:MM[:SS]]"
        raise configuration.ConfigurationException(fmt.format(string=p_string))

    hour = int(match.group(1))

    if match.group(3) is not None:
        minute = int(match.group(3))

    else:
        minute = 0

    if match.group(5) is not None:
        second = int(match.group(5))

    else:
        second = 0

    return datetime.time(hour=hour, minute=minute, second=second)

class TimingContext(object):

    def __init__(self, p_result_handler):
        self._result_handler = p_result_handler

    def __enter__(self):
        self._start = time.time()

    def __exit__(self, type, value, traceback):
        self._end = time.time()
        self._result_handler(self._end - self._start)



# -*- coding: utf-8 -*-

#    Copyright (C) 2019  Marcus Rickert
#
#    See https://github.com/marcus67/python_base_app
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

class InstallationException(Exception):
    pass


class SignalHangUp(Exception):
    pass


class ScriptExecutionError(Exception):

    def __init__(self, p_script_name, p_exit_code):
        super().__init__()
        self.script_name = p_script_name
        self.exit_code = p_exit_code


class ArtifactNotFoundException(Exception):

    def __init__(self, p_artifact_path, p_error_code, p_result_document=None):
        self._artifact_path = p_artifact_path
        self._error_code = p_error_code
        self._result_document = p_result_document

    def __str__(self):
        return "ArtifactNotFoundException with error code '%s' while accessing artifact '%s'" % (
            str(self._error_code), self._artifact_path)

    @property
    def result_document(self):
        return self._result_document


class ArtifactBlockedException(Exception):

    def __init__(self, p_artifact_path, p_error_code):
        self._artifact_path = p_artifact_path
        self._error_code = p_error_code

    def __str__(self):
        return "ArtifactBlockedException with error code '%s' while accessing artifact '%s'" % (
            str(self._error_code), self._artifact_path)


class TimeoutException(Exception):

    def __init__(self, p_artifact_path, p_error_code):
        self._artifact_path = p_artifact_path
        self._error_code = p_error_code

    def __str__(self):
        return "TimeOutException with error code '%s' while accessing artifact '%s'" % (
            str(self._error_code), self._artifact_path)


class UnauthorizedException(Exception):

    def __init__(self, p_artifact_path, p_login, p_error_code):
        self._artifact_path = p_artifact_path
        self._login = p_login
        self._error_code = p_error_code

    def __str__(self):
        return "UnauthorizedException with error code '%s' while accessing artifact '%s' using login '%s'" % (
            str(self._error_code), self._artifact_path, self._login)


class DuplicateKeyException(Exception):

    def __init__(self, p_key, p_error_code):
        self._key = p_key
        self._error_code = p_error_code

    def __str__(self):
        return "DuplicateKeyException with error '%s' for key '%s'" % (
            str(self._error_code), str(self._key))



# -*- coding: utf-8 -*-

# Copyright (C) 2019  Marcus Rickert
#
# See https://github.com/marcus67/python_base_app
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from python_base_app import log_handling

import abc


class BaseAudioPlayer(object, metaclass=abc.ABCMeta):

    def __init__(self):

        self._logger = log_handling.get_logger(self.__class__.__name__)

    @abc.abstractmethod
    def play_audio_file(self, p_audio_filename):  # pragma: no cover
        pass

    def stop(self):
        pass



# -*- coding: utf-8 -*-

# Copyright (C) 2019  Marcus Rickert
#
# See https://github.com/marcus67/little_brother
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

git_metadata = {
    "commit_id": "<unknown>",
    "branch": "<unknown>",
    "author_name": "<unknown>",
    "author_email": "<unknown>"
}

try:
    from python_base_app import git_metadata

    git_metadata = {
        "commit_id": git_metadata.GIT_COMMIT_ID[0:16],
        "branch": git_metadata.GIT_BRANCH,
        "author_name": git_metadata.GIT_AUTHOR_NAME,
        "author_email":git_metadata.GIT_AUTHOR_EMAIL
    }

except Exception as e:
    print(str(e))


# -*- coding: utf-8 -*-

#    Copyright (C) 2019  Marcus Rickert
#
#    See https://github.com/marcus67/python_base_app
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import wtforms.widgets

from python_base_app import tools


class BaseCustomField(wtforms.Field):
    extra_css_classes = ""


class DurationField(BaseCustomField):
    widget = wtforms.widgets.TextInput()

    def __init__(self, *args, **largs):

        super().__init__(*args, **largs)
        self.data = None
        self.invalid_data = None

    def _value(self):
        if self.invalid_data is not None:
            return self.invalid_data

        elif self.data is None:
            return '-'

        else:
            return tools.get_duration_as_string(p_seconds=self.data, p_include_seconds=False)

    def process_formdata(self, valuelist):
        if valuelist:
            try:
                self.data = tools.get_string_as_duration(p_string=valuelist[0])
                self.invalid_data = None

            except Exception as e:

                self.invalid_data = valuelist[0]
                raise wtforms.validators.ValidationError(message=str(e))
        else:
            self.data = None
            self.invalid_data = None


class TimeField(BaseCustomField):
    widget = wtforms.widgets.TextInput()

    def __init__(self, *args, **largs):

        super().__init__(*args, **largs)
        self.data = None
        self.invalid_data = None

    def _value(self):

        if self.invalid_data is not None:
            return self.invalid_data

        elif self.data is None:
            return '-'
        else:
            return tools.get_time_as_string(p_timestamp=self.data, p_include_seconds=False)

    def process_formdata(self, valuelist):
        if valuelist:

            try:
                self.data = tools.get_string_as_time(p_string=valuelist[0])

            except Exception as e:

                self.invalid_data = valuelist[0]
                raise wtforms.validators.ValidationError(message=str(e))

        else:
            self.data = None
            self.invalid_data = None


class BooleanField(BaseCustomField):
    widget = wtforms.widgets.CheckboxInput()

    extra_css_classes = "move-left"

    def __init__(self, *args, **largs):

        super().__init__(*args, **largs)
        self.data = None

    def _value(self):

        return self.data is not None and self.data

    def process_formdata(self, valuelist):
        if valuelist:

            try:
                self.data = 1 if valuelist[0] else 0

            except Exception as e:
                raise wtforms.validators.ValidationError(message=str(e))

        else:
            self.data = None



# -*- coding: utf-8 -*-

#    Copyright (C) 2019  Marcus Rickert
#
#    See https://github.com/marcus67/python_base_app
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import argparse
import datetime
import heapq
import os
import signal
import time
import pathlib

from python_base_app import configuration
from python_base_app import daemon
from python_base_app import exceptions
from python_base_app import log_handling
from python_base_app import tools

DEFAULT_DEBUG_MODE = False

DEFAULT_LOG_LEVEL = 'INFO'
DEFAULT_SPOOL_BASE_DIR = "/var/spool"
DEFAULT_USER_CONFIG_DIR = ".config"
DEFAULT_CONF_EXTENSION = ".conf"

TIME_SLACK = 0.1  # seconds
ETERNITY = 24 * 3600  # seconds
DEFAULT_TASK_INTERVAL = 10  # seconds
DEFAULT_MAXIMUM_TIMER_SLACK = 5  # second
DEFAULT_MINIMUM_DOWNTIME_DURATION = 20  # seconds


class BaseAppConfigModel(configuration.ConfigModel):

    def __init__(self, p_section_name="BaseApp"):
        super().__init__(p_section_name=p_section_name)

        self.debug_mode = DEFAULT_DEBUG_MODE
        self.log_level = configuration.NONE_STRING
        self.spool_dir = configuration.NONE_STRING
        self.minimum_downtime_duration = DEFAULT_MINIMUM_DOWNTIME_DURATION
        self.maximum_timer_slack = DEFAULT_MAXIMUM_TIMER_SLACK

    def is_active(self):
        return True


class RecurringTask(object):

    def __init__(self, p_name, p_handler_method, p_interval=DEFAULT_TASK_INTERVAL, p_fixed_schedule=False):

        self.name = p_name
        self.handler_method = p_handler_method
        self.interval = p_interval
        self.next_execution = None
        self.fixed_schedule = p_fixed_schedule

    def __lt__(self, p_other):
        return self.next_execution < p_other

    def __gt__(self, p_other):
        return self.next_execution > p_other

    def __sub__(self, p_other):
        if isinstance(p_other, RecurringTask):
            return self.next_execution - p_other.next_execution

        else:
            return self.next_execution - p_other

    def __rsub__(self, p_other):
        if isinstance(p_other, RecurringTask):
            return p_other.next_execution - self.next_execution

        else:
            return p_other - self.next_execution

    def get_heap_entry(self):

        # return (self, self)
        return self

    def compute_next_execution_time(self):

        if self.next_execution is None:
            self.next_execution = datetime.datetime.utcnow()

        elif self.fixed_schedule:
            self.next_execution = self.next_execution + datetime.timedelta(seconds=self.interval)

        else:
            self.next_execution = datetime.datetime.utcnow() + datetime.timedelta(seconds=self.interval)

    def adapt_to_delay(self, p_delay):

        if self.next_execution is not None:
            self.next_execution = self.next_execution + datetime.timedelta(seconds=p_delay)


class BaseApp(daemon.Daemon):

    def __init__(self, p_app_name, p_pid_file, p_arguments, p_dir_name):

        super().__init__(pidfile=p_pid_file)

        self._app_name = p_app_name
        self._dir_name = p_dir_name
        self._arguments = p_arguments
        self._logger = log_handling.get_logger(self.__class__.__name__)
        self._config = None
        self._recurring_tasks = []
        self._downtime = 0

        # Only temporary until the app has been initialized completely!
        self._app_config = BaseAppConfigModel()

    def check_user_configuration_file(self):

        home = str(pathlib.Path.home())
        user_configuration_file = os.path.join(home, DEFAULT_USER_CONFIG_DIR, self._app_name) + DEFAULT_CONF_EXTENSION

        fmt = "Looking for user configuration file in {path}"
        self._logger.info(fmt.format(path=user_configuration_file))

        if os.path.exists(user_configuration_file):
            self._arguments.configurations.append(user_configuration_file)

    @property
    def down_time(self):
        return self._downtime

    def reset_down_time(self):
        self._downtime = 0

    def add_recurring_task(self, p_recurring_task):

        p_recurring_task.compute_next_execution_time()
        heapq.heappush(self._recurring_tasks, p_recurring_task.get_heap_entry())

    @staticmethod
    def configuration_factory():
        return configuration.Configuration()

    def load_configuration(self, p_configuration):

        self._app_config = p_configuration[self._app_name]

        for afile in self._arguments.configurations:
            p_configuration.read_config_file(afile)

        p_configuration.read_environment_parameters(p_environment_dict=os.environ)
        p_configuration.read_command_line_parameters(p_parameters=self._arguments.cmd_line_options)

        if self._app_config.log_level is not None:
            log_handling.set_level(self._app_config.log_level)

        if self._app_config.spool_dir is None:
            self._app_config.spool_dir = os.path.join(DEFAULT_SPOOL_BASE_DIR, self._dir_name)

        return p_configuration

    def check_configuration(self):

        logger = log_handling.get_logger()

        self.load_configuration(self.configuration_factory())

        fmt = "%d configuration files are Ok!" % len(self._arguments.configurations)
        logger.info(fmt)

    def handle_sighup(self, p_signum, p_stackframe):

        fmt = "Received signal %d" % p_signum
        _ = p_stackframe
        self._logger.info(fmt)

        raise exceptions.SignalHangUp()

    def install_handlers(self):

        signal.signal(signal.SIGTERM, self.handle_sighup)
        signal.signal(signal.SIGHUP, self.handle_sighup)
        signal.signal(signal.SIGINT, self.handle_sighup)  # CTRL-C
        signal.pthread_sigmask(signal.SIG_UNBLOCK, [signal.SIGTERM, signal.SIGHUP, signal.SIGINT])

    def prepare_services(self, p_full_startup=True):

        pass

    def start_services(self):

        pass

    def adapt_active_recurring_tasks(self, p_delay):

        # for (_next_execution, task) in self._recurring_tasks:
        #    task.adapt_to_delay(p_delay=p_delay)
        for task in self._recurring_tasks:
            task.adapt_to_delay(p_delay=p_delay)

        heapq.heapify(self._recurring_tasks)

    def stop_event_queue(self):
        self._done = True

    def event_queue(self):

        self._done = False

        while not self._done:
            try:
                now = datetime.datetime.utcnow()

                if len(self._recurring_tasks) > 0:
                    task = self._recurring_tasks[0]
                    wait_in_seconds = (task - now).total_seconds()

                else:
                    task = None
                    wait_in_seconds = ETERNITY

                if wait_in_seconds > 0:
                    try:
                        fmt = "Sleeping for {seconds} seconds (or until next signal)"
                        self._logger.debug(fmt.format(seconds=wait_in_seconds))

                        signal.pthread_sigmask(signal.SIG_UNBLOCK, [signal.SIGTERM, signal.SIGHUP])
                        time.sleep(wait_in_seconds)

                    except exceptions.SignalHangUp as e:
                        raise e

                    except Exception as e:
                        if self._app_config.debug_mode:
                            fmt = "Propagating exception due to debug_mode=True"
                            self._logger.warn(fmt)
                            raise e

                        fmt = "Exception %s while waiting for signal" % str(e)
                        self._logger.error(fmt)

                    fmt = "Woken by signal"
                    self._logger.debug(fmt)

                    if task is not None:
                        now = datetime.datetime.utcnow()
                        overslept_in_seconds = (now - task).total_seconds()

                        if overslept_in_seconds > self._app_config.maximum_timer_slack:
                            self.track_downtime(p_downtime=overslept_in_seconds)

                if len(self._recurring_tasks) > 0:
                    task_executed = True

                    while task_executed:
                        task = self._recurring_tasks[0]
                        now = datetime.datetime.utcnow()

                        if now > task:
                            delay = (now - task).total_seconds()

                            task = heapq.heappop(self._recurring_tasks)
                            self.add_recurring_task(p_recurring_task=task)

                            fmt = "Executing task {task} {secs:.3f} [s] behind schedule... *** START ***"
                            self._logger.debug(fmt.format(task=task.name, secs=delay))
                            task.handler_method()
                            fmt = "Executing task {task} {secs:.3f} [s] behind schedule... *** END ***"
                            self._logger.debug(fmt.format(task=task.name, secs=delay))

                            if delay > self._app_config.minimum_downtime_duration:
                                self.track_downtime(p_downtime=delay)

                        else:
                            task_executed = False

                    if self._downtime > 0:
                        self.handle_downtime(p_downtime=int(self._downtime))
                        self.reset_down_time()

            except exceptions.SignalHangUp:
                fmt = "Event queue interrupted by signal"
                self._logger.info(fmt)

                self._done = True

            except Exception as e:
                if self._app_config.debug_mode:
                    fmt = "Propagating exception due to debug_mode=True"
                    self._logger.warn(fmt)
                    raise e

                fmt = "Exception %s in event queue" % str(e)
                self._logger.error(fmt)
                tools.log_stack_trace(p_logger=self._logger)

            if self._arguments.single_run:
                self._done = True

    def track_downtime(self, p_downtime):

        fmt = "Detected delay of {seconds} seconds -> adding to downtime timer"
        self._logger.info(fmt.format(seconds=p_downtime))

        self._downtime += p_downtime
        self.adapt_active_recurring_tasks(p_delay=p_downtime)

    def handle_downtime(self, p_downtime):

        fmt = "Accumulated downtime of {seconds} seconds ignored."
        self._logger.warning(fmt.format(seconds=p_downtime))

    def stop_services(self):

        pass

    def basic_init(self, p_full_startup=True):

        try:
            self._config = self.load_configuration(self.configuration_factory())
            self.prepare_services(p_full_startup=p_full_startup)

        except Exception as e:
            fmt = "Error '{msg}' in basic_init()"
            self._logger.error(fmt.format(msg=str(e)))

            # if self._app_config.debug_mode:
            #     fmt = "Propagating exception due to debug_mode=True"
            #     self._logger.warn(fmt)
            #     raise e

            raise e

    def run(self):

        previous_exception = None

        fmt = "Starting app '%s'" % self._app_name
        self._logger.info(fmt)

        try:
            self.basic_init()
            self.install_handlers()
            self.start_services()
            self.event_queue()

        except Exception as e:
            fmt = "Exception '%s' in run()" % str(e)
            self._logger.error(fmt)

            previous_exception = e

        finally:
            try:
                self.stop_services()

            except Exception as e:
                fmt = "Exception '%s' while stopping services" % str(e)
                self._logger.error(fmt)

        if previous_exception is not None:
            raise previous_exception

        fmt = "Terminating app '%s'" % self._app_name
        self._logger.info(fmt)

    def run_special_commands(self, p_arguments):

        _arguments = p_arguments
        return False


def get_argument_parser(p_app_name):
    parser = argparse.ArgumentParser()
    parser.add_argument('--config', nargs='*', dest='configurations', default=[],
                        help='file names of the configuration files')
    parser.add_argument('--option', nargs='*', dest='cmd_line_options', default=[],
                        help='Additional configuration settings formatted as '
                             'SECTION.OPTION=VALUE (overriding settings in configuration files)')
    parser.add_argument('--pidfile', dest='pid_file',
                        help='name of the PID file', default='/var/run/%s/%s.pid' % (p_app_name, p_app_name))
    parser.add_argument('--logdir', dest='log_dir', default=None,
                        help='base path for logging files')
    parser.add_argument('--loglevel', dest='log_level', default=DEFAULT_LOG_LEVEL,
                        help='logging level', choices=['WARN', 'INFO', 'DEBUG'])
    parser.add_argument('--application-owner', dest='app_owner', default=None,
                        help='name of user running application')
    parser.add_argument('--daemonize', dest='daemonize', action='store_const', const=True, default=False,
                        help='start app as daemon process')
    parser.add_argument('--check-installation', dest='check_installation', action='store_const', const=True,
                        default=False,
                        help='Tests the existence of required directories and files as well as their access rights')
    parser.add_argument('--check-configuration', dest='check_configuration', action='store_const', const=True,
                        default=False,
                        help='Validates the configurations files')
    parser.add_argument('--kill', dest='kill', action='store_const', const=True, default=False,
                        help='Terminates the running daemon process')
    parser.add_argument('--single-run', dest='single_run', action='store_const', const=True, default=False,
                        help='Executes a single run of the application and exists')

    return parser


def check_installation(p_arguments):
    pid_file_directory = os.path.dirname(p_arguments.pid_file)

    if p_arguments.log_dir is not None:
        tools.test_mode(p_arguments.log_dir, p_app_owner=p_arguments.app_owner, p_is_directory=True, p_executable=True,
                        p_writable=True, p_other_access=False)

    if p_arguments.daemonize:
        tools.test_mode(pid_file_directory, p_app_owner=p_arguments.app_owner, p_is_directory=True, p_executable=True,
                        p_writable=True)

    for config_file in p_arguments.configurations:
        tools.test_mode(config_file, p_app_owner=p_arguments.app_owner, p_other_access=False)

    logger = log_handling.get_logger()
    logger.info("Installation OK!")


def main(p_app_name, p_app_class, p_argument_parser):
    process_result = 0
    logger = log_handling.get_logger()
    arguments = p_argument_parser.parse_args()

    try:
        if arguments.daemonize and arguments.log_dir is None:
            raise configuration.ConfigurationException("Option --daemonize requires option --logdir!")

        default_log_file = '%s.log' % p_app_name

        log_handling.start_logging(p_level=arguments.log_level, p_log_dir=arguments.log_dir,
                                   p_log_file=default_log_file)
        logger = log_handling.get_logger()

        if arguments.check_installation:
            logger.info("Checking installation...")
            check_installation(p_arguments=arguments)

        else:

            app = p_app_class(p_pid_file=arguments.pid_file, p_arguments=arguments, p_app_name=p_app_name)

            if len(arguments.configurations) == 0:
                logger.warning("No configuration files specified")

            if arguments.kill:
                logger.info("Killing active daemon process...")
                app.stop()

            elif arguments.check_configuration:
                logger.info("Checking configuration files...")
                app.check_configuration()

            elif arguments.daemonize:
                logger.info("Starting daemon process...")
                app.start()

            elif not app.run_special_commands(p_arguments=arguments):
                logger.info("Starting as a normal foreground process...")
                app.run()

    except configuration.ConfigurationException as e:
        logger.error(str(e))
        process_result = 3

    except exceptions.InstallationException as e:
        logger.error(str(e))
        process_result = 2

    except Exception as e:

        tools.handle_fatal_exception(p_exception=e, p_logger=logger)
        tools.log_stack_trace(p_logger=logger)
        process_result = 1

    fmt = 'Terminated with exit code %d' % process_result
    logger.info(fmt)

    return process_result



# -*- coding: utf-8 -*-

#    Copyright (C) 2019  Marcus Rickert
#
#    See https://github.com/marcus67/python_base_app
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import abc
import configparser
import re

from python_base_app import log_handling
from python_base_app import tools

REGEX_CMDLINE_PARAMETER = re.compile(r"([-a-zA-Z_0-9]+)\.([a-zA-Z_0-9]+)=(.*)")
REGEX_ENV_PARAMETER = re.compile("([a-zA-Z_0-9]+)__([a-zA-Z_0-9]+)")

NONE_BOOLEAN = type(True)
NONE_INTEGER = type(1)
NONE_STRING = type("X")

OPTION_ARRAY_PATTERN = re.compile(r"([^[]*)\[([0-9]+)\]")

VALID_BOOLEAN_TRUE_VALUES = ['1', 'TRUE', 'T', 'YES', 'WAHR', 'JA', 'J']
VALID_BOOLEAN_FALSE_VALUES = ['0', 'FALSE', 'F', 'NO', 'FALSCH', 'NEIN', 'N']


class ConfigurationException(Exception):

    def __init__(self, p_text):
        super(ConfigurationException, self).__init__(p_text)


class ConfigurationSectionHandler(object, metaclass=abc.ABCMeta):

    def __init__(self, p_section_prefix):
        self._section_prefix = p_section_prefix
        self._configuration = None

    @property
    def section_prefix(self):
        return self._section_prefix

    @abc.abstractmethod
    def handle_section(self, p_section_name):
        pass

    def set_configuration(self, p_configuration):
        self._configuration = p_configuration

    def scan(self, p_section):
        self._configuration.add_section(p_section=p_section)
        self._configuration.scan_section(p_section_name=p_section.section_name)


NONE_TYPE_PREFIX = "_TYPE_"
NONE_ARRAY_TYPE_PREFIX = "_ARRAY_TYPE_"


class SimpleConfigurationSectionHandler(ConfigurationSectionHandler):

    def __init__(self, p_config_model):
        super().__init__(p_section_prefix=p_config_model._section_name)
        self._config_model = p_config_model

    def handle_section(self, p_section_name):
        self.scan(p_section=self._config_model)


class ConfigModel(object):

    def __init__(self, p_section_name, p_class_name=None):

        self.section_name = p_section_name
        self._options = {}

    def is_active(self):
        raise NotImplementedError("ConfigModel.is_active")

    def get_option_type(self, p_option_name):

        p_effective_name = NONE_ARRAY_TYPE_PREFIX + p_option_name

        value = self.__dict__.get(p_effective_name)

        if value is not None:
            return "list_" + value.__name__

        else:
            p_effective_name = NONE_TYPE_PREFIX + p_option_name

            value = self.__dict__.get(p_effective_name)

            if value is not None:
                return value.__name__

            else:
                value = self.__dict__[p_option_name]

                if isinstance(value, list):
                    if len(value) == 0:
                        fmt = "Option '{option}' defines empty array without type"
                        raise ConfigurationException(fmt.format(option=p_option_name))

                    return "list_" + type(value[0]).__name__

                else:
                    return type(value).__name__

    def has_option(self, p_option_name):

        p_effective_name = NONE_TYPE_PREFIX + p_option_name
        p_effective_name_list = NONE_ARRAY_TYPE_PREFIX + p_option_name

        return p_option_name in self.__dict__ or \
               p_effective_name in self.__dict__ or p_effective_name_list in self.__dict__

    def __getattr__(self, p_option_name):

        # Note: __getattr__ is ONLY called if the builtin mechanism did not find the attribute, that is
        # for all existing attributes the method WILL NEVER BE CALLED!

        p_effective_name = NONE_TYPE_PREFIX + p_option_name
        value = self.__dict__.get(p_effective_name)

        if value is not None:
            return None

        else:
            p_effective_name = NONE_ARRAY_TYPE_PREFIX + p_option_name
            value = self.__dict__.get(p_effective_name)

            if value is not None:
                return []

            else:
                fmt = "unknown option name '{name}'"
                raise AttributeError(fmt.format(name=p_option_name))

    def __setattr__(self, p_option_name, p_value):

        if isinstance(p_value, list):
            if len(p_value) == 1 and isinstance(p_value[0], type):
                p_effective_name = NONE_ARRAY_TYPE_PREFIX + p_option_name
                self.__dict__[p_effective_name] = p_value[0]

            else:
                self.__dict__[p_option_name] = p_value

        elif isinstance(p_value, type):
            p_effective_name = NONE_TYPE_PREFIX + p_option_name
            self.__dict__[p_effective_name] = p_value

        else:
            self.__dict__[p_option_name] = p_value

    def post_process(self):

        pass


class Configuration(ConfigModel):

    def __init__(self):

        super().__init__(p_section_name="_Configuration_")

        self._sections = {}
        self._logger = log_handling.get_logger(self.__class__.__name__)
        self._section_handlers = []

    def add_section(self, p_section):

        if p_section.section_name in self._sections:
            fmt = "Overwriting existing section '%s'" % p_section.section_name
            self._logger.warning(fmt)

        self._sections[p_section.section_name] = p_section

    def register_section_handler(self, p_section_handler):

        self._section_handlers.append(p_section_handler)
        p_section_handler.set_configuration(p_configuration=self)

    def __getitem__(self, p_key):

        if p_key not in self._sections:
            raise ConfigurationException("No section '%s' configured in class %s" % (p_key, self.__class__.__name__))

        return self._sections[p_key]

    def set_config_value(self, p_section_name, p_option, p_option_value):

        section = self._sections.get(p_section_name)
        append_to_list = False

        if section is None:
            raise ConfigurationException("Invalid section name '%s'" % p_section_name)

        match = OPTION_ARRAY_PATTERN.match(p_option)

        if match is not None:
            if int(match.group(2)) > 0:
                append_to_list = True
            p_option = match.group(1)

        if not section.has_option(p_option_name=p_option):
            raise ConfigurationException(
                "Configuration file contains invalid setting '%s' in section '%s'" % (p_option, p_section_name))

        option_type = section.get_option_type(p_option_name=p_option)
        upper_value = p_option_value.upper()

        if 'bool' in option_type:
            if upper_value in VALID_BOOLEAN_TRUE_VALUES:
                value = True

            elif upper_value in VALID_BOOLEAN_FALSE_VALUES:
                value = False

            else:
                raise ConfigurationException("Invalid Boolean value '%s' in setting '%s' of section '%s'" % (
                    p_option_value, p_option, p_section_name))

        elif 'int' in option_type:
            try:
                value = int(p_option_value)

            except Exception as e:
                raise ConfigurationException("Invalid numerical value '%s' in setting '%s' of section '%s': %s" % (
                    p_option_value, p_option, p_section_name, str(e)))

        else:
            value = p_option_value

        if 'list' in option_type:
            if append_to_list:
                getattr(section, p_option).append(value)

            else:
                setattr(section, p_option, [value])

        else:
            setattr(section, p_option, value)

    def scan_section(self, p_section_name):

        section = self._sections.get(p_section_name)

        fmt = "Scanning settings for section '%s'" % p_section_name
        self._logger.debug(fmt)

        if section is None:
            raise ConfigurationException("Invalid section name '%s'" % p_section_name)

        for option in self.config.options(p_section_name):
            option_value = self.config.get(p_section_name, option)
            self.set_config_value(
                p_section_name=p_section_name,
                p_option=option,
                p_option_value=option_value)

    def handle_section(self, p_section_name, p_ignore_invalid_sections=False, p_warn_about_invalid_sections=False):

        for section_handler in self._section_handlers:
            if p_section_name.startswith(section_handler.section_prefix):
                section_handler.handle_section(p_section_name=p_section_name)
                return

        if p_ignore_invalid_sections:
            if p_warn_about_invalid_sections:
                fmt = "Ignoring invalid section '%s' (not all sections registered?)" % p_section_name
                self._logger.warning(fmt)

        else:
            fmt = "Invalid section '%s'" % p_section_name
            self._logger.error(fmt)
            raise ConfigurationException("Configuration file contains invalid section '%s'" % p_section_name)

    def read_config_file(self, p_filename=None, p_config_string=None,
                         p_ignore_invalid_sections=False, p_warn_about_invalid_sections=False):

        errorMessage = None

        if p_filename is not None:
            fmt = "Reading configuration file from '%s'" % p_filename
            self._logger.info(fmt)

            self.config = configparser.ConfigParser(strict=False)
            self.config.optionxform = str  # make options case sensitive

            try:
                filesRead = self.config.read([p_filename], encoding="UTF-8")
                if len(filesRead) != 1:
                    fmt = "Error while reading configuration file '{filename}' (file probably does not exist)"
                    errorMessage = fmt.format(filename=p_filename)

            except Exception as e:
                fmt = "Exception '{msg}' while reading configuration file '{filename}'"
                errorMessage = fmt.format(msg=str(e), filename=p_filename)

        if p_config_string is not None:

            try:
                self.config = configparser.ConfigParser()
                self.config.read_string(p_config_string)

            except Exception as e:
                fmt = "Exception '{msg}' while reading setting"
                errorMessage = fmt.format(msg=str(e))

        if errorMessage is not None:
            raise ConfigurationException(errorMessage)

        for section_name in self.config.sections():
            if section_name in self._sections:

                new_section = self._sections[section_name]
                setattr(self, section_name, new_section)
                self.scan_section(section_name)

            else:

                self.handle_section(p_section_name=section_name,
                                    p_ignore_invalid_sections=p_ignore_invalid_sections,
                                    p_warn_about_invalid_sections=p_warn_about_invalid_sections)

    def read_command_line_parameters(self, p_parameters):

        for par in p_parameters:
            result = REGEX_CMDLINE_PARAMETER.match(par)

            if result:

                section_name = result.group(1)
                option_name = result.group(2)
                value = result.group(3)

                protected_value = tools.protect_password_value(p_name=option_name, p_value=value)

                fmt = "Command line setting: set '[%s]%s' to value '%s'" % (
                    section_name, option_name, protected_value)
                self._logger.info(fmt)

                self.set_config_value(
                    p_section_name=section_name,
                    p_option=option_name,
                    p_option_value=value)

            else:
                fmt = "Incorrectly formatted command line setting: %s" % par
                self._logger.warning(fmt)

    def read_environment_parameters(self, p_environment_dict):

        for (name, value) in dict(p_environment_dict).items():
            result = REGEX_ENV_PARAMETER.match(name)

            if result:
                section_name = result.group(1)
                option_name = result.group(2)

                protected_value = tools.protect_password_value(p_name=option_name, p_value=value)

                fmt = "Environment setting: set '[{section_name}]{option_name}' to value '{value}'"
                self._logger.info(fmt.format(section_name=section_name, option_name=option_name, value=protected_value))

                self.set_config_value(
                    p_section_name=section_name,
                    p_option=option_name,
                    p_option_value=value)



# -*- coding: utf-8 -*-

# Copyright (C) 2019  Marcus Rickert
#
# See https://github.com/marcus67/python_base_app
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os.path
import shlex
import subprocess

from python_base_app import configuration
from python_base_app import base_audio_player

class Mpg123AudioPlayer(base_audio_player.BaseAudioPlayer):


    def __init__(self, p_mpg123_binary):

        super().__init__()
        self._mpg123_binary = p_mpg123_binary

        if not os.path.exists(p_mpg123_binary):
            fmt = "Cannot find mpg123 binary at path '{path}'"
            raise configuration.ConfigurationException(fmt.format(path=p_mpg123_binary))

        fmt = "using audio player '{path}'"
        self._logger.info(fmt.format(path=p_mpg123_binary))


    def play_audio_file(self, p_audio_filename):  # pragma: no cover
        play_command = "{binary} {file}".format(binary=self._mpg123_binary, file=p_audio_filename)

        cmd_array = shlex.split(play_command)
        subprocess.run(cmd_array)



# -*- coding: utf-8 -*-

#    Copyright (C) 2019  Marcus Rickert
#
#    See https://github.com/marcus67/python_base_app
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from python_base_app import tools


class ViewInfo(object):

    def __init__(self, p_html_key, p_parent=None):

        self.parent = p_parent
        self._html_key = tools.get_safe_attribute_name(p_string=p_html_key)

    @property
    def html_key(self):

        if self.parent is None:
            return self._html_key

        else:
            return "%s_%s" % (self.parent.html_key, self._html_key)



# -*- coding: utf-8 -*-

#    Copyright (C) 2019  Marcus Rickert
#
#    See https://github.com/marcus67/python_base_app
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import flask

from flask_helpers import blueprint_adapter

import python_base_app

ACTUATOR_BLUEPRINT_NAME = "actuator"
ACTUATOR_BLUEPRINT_ADAPTER = blueprint_adapter.BlueprintAdapter()

class ActuatorViewHandler(object):

    def __init__(self, p_app, p_url_prefix):
        self._blueprint = flask.Blueprint(ACTUATOR_BLUEPRINT_NAME, python_base_app.__name__)
        ACTUATOR_BLUEPRINT_ADAPTER.assign_view_handler_instance(p_blueprint=self._blueprint,
            p_view_handler_instance=self)
        ACTUATOR_BLUEPRINT_ADAPTER.check_view_methods()
        p_app.register_blueprint(self._blueprint, url_prefix=p_url_prefix)

    @ACTUATOR_BLUEPRINT_ADAPTER.route_method(p_rule="/health")
    def health(self):

        return flask.Response("ok")

    def destroy(self):
        ACTUATOR_BLUEPRINT_ADAPTER.unassign_view_handler_instances()


#    Copyright (C) 2019  Marcus Rickert
#
#    See https://github.com/marcus67/python_base_app
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from python_base_app.test import base_test


class TestPytest(base_test.BaseTestCase):

    def test_pytest(self):
        self.execute_pytest()



# -*- coding: utf-8 -*-

# Copyright (C) 2019  Marcus Rickert
#
# See https://github.com/marcus67/little_brother
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import datetime
import hashlib
import locale
import os.path
import shlex
import subprocess

from python_base_app import configuration
from python_base_app import mpg123_audio_player
from python_base_app import notification_handler
from python_base_app import pyglet_audio_player

DEFAULT_SPEECH_GENERATOR_CMD_LINE = '/usr/bin/festival --tts --language american_english {{{pattern}}}'.format(
    pattern=notification_handler.REPLACE_PATTERN_AUDIO_TEXT_FILENAME)
DEFAULT_AUDIO_MIXER_BIN = '/usr/bin/amixer'
DEFAULT_AUDIO_FILE_PREFIX = "python-base-app-speech-"
DEFAULT_SPOOL_DIRECTORY = "/tmp"
DEFAULT_SPEECH_WORDS_PER_MINUTE = 100

SECTION_NAME = "AudioHandler"

SPEECH_ENGINE_GOOGLE = "google"
SPEECH_ENGINE_EXTERNAL = "external"

AUDIO_TEXT_FILE = "audio.txt"

AUDIO_PLAYER_PYGLET = "pyglet"
AUDIO_PLAYER_MPG123 = "mpg123"

DEFAULT_MPG123_binary = "/usr/bin/mpg123"

DEFAULT_AUDIO_PLAYER = AUDIO_PLAYER_MPG123


class AudioHandlerConfigModel(notification_handler.NotificationHandlerConfigModel):

    def __init__(self):
        super().__init__(p_section_name=SECTION_NAME)

        self.cache_audio_files = False
        self.speech_engine = configuration.NONE_STRING
        self.speech_words_per_minute = DEFAULT_SPEECH_WORDS_PER_MINUTE
        self.spool_dir = DEFAULT_SPOOL_DIRECTORY
        self.audio_file_prefix = DEFAULT_AUDIO_FILE_PREFIX
        self.audio_mixer_bin = DEFAULT_AUDIO_MIXER_BIN
        self.audio_mixer_volume = configuration.NONE_INTEGER  # in percent
        self.speech_generator_cmd_line = DEFAULT_SPEECH_GENERATOR_CMD_LINE
        self.audio_player = DEFAULT_AUDIO_PLAYER
        self.mpg123_binary = DEFAULT_MPG123_binary

    def is_active(self):
        return self.speech_engine is not None


class AudioHandler(notification_handler.NotificationHandler):

    def __init__(self, p_config):

        self._google_speak = None
        self._audio_player = None

        super().__init__(p_config=p_config)

    # def init_engine_pyttsx3(self):
    #
    #     self._pyttsx3_engine = pyttsx3.init()
    #     self._pyttsx3_engine.setProperty('rate', self._config.speech_words_per_minute)

    def check_audio_player(self):

        try:
            # if self._config.audio_player == AUDIO_PLAYER_PLAYSOUND:
            #     self._audio_player = playsound_audio_player.PlaysoundAudioPlayer()

            if self._config.audio_player == AUDIO_PLAYER_PYGLET:
                self._audio_player = pyglet_audio_player.PygletAudioPlayer()

            elif self._config.audio_player == AUDIO_PLAYER_MPG123:
                self._audio_player = mpg123_audio_player.Mpg123AudioPlayer(self._config.mpg123_binary)

            else:
                fmt = "Invalid audio player '{player}'"
                self._logger.error(fmt.format(player=self._config.audio_player))
                raise configuration.ConfigurationException(fmt)

        except Exception:
            fmt = "Cannot load audio player '{player}'"
            self._logger.error(fmt.format(player=self._config.audio_player))
            raise configuration.ConfigurationException(fmt)

    def init_engine_google(self):

        try:
            import python_google_speak.speech_generator
            self._google_speak = python_google_speak.speech_generator

        except:
            fmt = "init_engine_google(): cannot load module 'python_google_speak'"
            self._logger.error(fmt)
            raise configuration.ConfigurationException(fmt)

        self.check_audio_player()

    def init_engine(self):

        if self._config.speech_engine == SPEECH_ENGINE_EXTERNAL:
            pass

        # elif self._config.speech_engine == SPEECH_ENGINE_PYTTSX3:
        #     self.init_engine_pyttsx3()

        elif self._config.speech_engine == SPEECH_ENGINE_GOOGLE:
            self.init_engine_google()

        else:
            fmt = "init_engine(): invalid speech engine '%s'" % self._config.speech_engine
            self._logger.error(fmt)
            raise configuration.ConfigurationException(fmt)

    def stop_engine(self):
        if self._audio_player is not None:
            self._audio_player.stop()

        super().stop_engine()

    def get_audio_filename(self, p_text, p_locale=None):

        if p_locale is None:
            p_locale = locale.getdefaultlocale()[0]

        text = (p_text + p_locale).encode("UTF-8")
        hash_value = hashlib.sha224(text).hexdigest()
        return os.path.join(self._config.spool_dir, self._config.audio_file_prefix + hash_value)

    def _notify(self, p_text, p_locale=None):

        if self._config.speech_engine == SPEECH_ENGINE_EXTERNAL:
            self.speak_external_command(p_text=p_text, p_locale=p_locale)

        # elif self._config.speech_engine == SPEECH_ENGINE_PYTTSX3:
        #     self.speak_pyttsx(p_text=p_text, p_locale=p_locale)

        elif self._config.speech_engine == SPEECH_ENGINE_GOOGLE:
            self.speak_google(p_text=p_text, p_locale=p_locale)

        else:
            fmt = "_notify(): invalid speech engine '%s'" % self._config.speech_engine
            self._logger.error(fmt)
            raise configuration.ConfigurationException(fmt)

        self._recent_texts[p_text] = datetime.datetime.now()

    def speak_google(self, p_text, p_locale=None):

        self.set_volume()

        audio_filename = self.get_audio_filename(p_text=p_text, p_locale=p_locale)

        if not os.path.exists(path=audio_filename):
            google_speak = self._google_speak.SpeechGenerator(p_locale=p_locale)
            data = google_speak.generate_audio_data(p_text=p_text)

            with open(audio_filename, "wb") as f:
                f.write(data)

        self._audio_player.play_audio_file(audio_filename)

    def set_volume(self):

        if self._config.audio_mixer_volume is not None:
            cmd_line = "%s -q sset Master %d%%" % (
                shlex.quote(self._config.audio_mixer_bin),
                self._config.audio_mixer_volume)

            try:
                fmt = "speak_external_command(): set volume to %d%%" % self._config.audio_mixer_volume
                self._logger.debug(fmt)

                subprocess.run(cmd_line, shell=True)

            except Exception as e:

                fmt = "cannot set audio volume using '%s': exception %s" % (cmd_line, str(e))
                self._logger.warning(fmt)

    def speak_external_command(self, p_text, p_locale=None):

        self.set_volume()

        audio_text_filename = "<UNDEFINED>"

        if notification_handler.REPLACE_PATTERN_AUDIO_TEXT_FILENAME in self._config.speech_generator_cmd_line:
            audio_text_filename = os.path.join(self._config.spool_dir, AUDIO_TEXT_FILE)

            try:
                with open(audio_text_filename, "w") as f:
                    f.write(p_text)

            except Exception as e:
                fmt = "exception '{estr}' while writing text '{text}' into audio playback file {filename}"
                self._logger.warning(fmt.format(estr=str(e), text=p_text, filename=audio_text_filename))
                return

        replacements = {
            notification_handler.REPLACE_PATTERN_AUDIO_TEXT_FILENAME: audio_text_filename,
            notification_handler.REPLACE_PATTERN_AUDIO_TEXT: p_text
        }

        cmd_line = self._config.speech_generator_cmd_line.format(**replacements)

        try:

            fmt = "speak_external_command(): execute '%s'" % cmd_line
            self._logger.debug(fmt)
            subprocess.run(cmd_line, shell=True)

        except Exception as e:

            fmt = "cannot playback text '%s': exception %s" % (p_text, str(e))
            self._logger.warning(fmt)



# -*- coding: utf-8 -*-

#    Copyright (C) 2019  Marcus Rickert
#
#    See https://github.com/marcus67/python_base_app
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from setuptools import setup
from os import path

this_directory = path.abspath(path.dirname(__file__))

with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

with open(path.join(this_directory, 'requirements.txt')) as f:
    install_requires = f.read().splitlines()

setup_params = {
    # standard setup configuration
    "name" : "python-base-app",
    "version" : "0.1.8",
    "description" : "Generic tools and base classes for Python applications with web interface.",
    "author" : "Marcus Rickert",
    "author_email" : "marcus.rickert@web.de",
    "url" : "https://github.com/marcus67/python_base_app",
    
    "install_requires" : install_requires,

    "scripts": [
        "run_python_base_app_test_suite.py",
    ],

    "packages" : [ 'python_base_app', 'python_base_app.test' ],
    "include_package_data": True,
    
    "long_description" : long_description,
    "long_description_content_type" : 'text/markdown',
}

extended_setup_params = {
    # additional setup configuration used by CI stages
    "id": "python-base-app",
    "build_debian_package": False,
    "build_pypi_package": True,
    "publish_pypi_package": ['release'],
}
extended_setup_params.update(setup_params)

if __name__ == '__main__':
    setup(**setup_params)



#    Copyright (C) 2019  Marcus Rickert
#
#    See https://github.com/marcus67/python_base_app
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os

from python_base_app import audio_handler
from python_base_app import mpg123_audio_player
from python_base_app.test import base_test

HELLO_MPG = os.path.join(os.path.dirname(__file__), "resources/hello.mpg")


class TestAudioPlayer(base_test.BaseTestCase):

    @base_test.skip_if_env("NO_AUDIO_OUTPUT")
    def test_mpg123_player(self):

        a_config = audio_handler.AudioHandlerConfigModel()
        a_player = mpg123_audio_player.Mpg123AudioPlayer(p_mpg123_binary = a_config.mpg123_binary)
        a_player.play_audio_file(HELLO_MPG)



# -*- coding: utf-8 -*-

#    Copyright (C) 2019  Marcus Rickert
#
#    See https://github.com/marcus67/python_base_app
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License along
#    with this program; if not, write to the Free Software Foundation, Inc.,
#    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import logging
import threading
import time
import urllib.parse
from os.path import join

import flask.globals
import flask.wrappers
import flask_login

import flask_helpers

from python_base_app import actuator
from python_base_app import auth_view_handler
from python_base_app import configuration
from python_base_app import log_handling
from python_base_app import tools

DEFAULT_BASE_URL = ''
DEFAULT_INTERNAL_BASE_URL = ''

_ = lambda x: x


class BaseWebServerConfigModel(configuration.ConfigModel):

    def __init__(self, p_section_name):
        super(BaseWebServerConfigModel, self).__init__(p_section_name)
        self.scheme = "http"
        self.host = "0.0.0.0"
        self.port = configuration.NONE_INTEGER
        self.base_url = DEFAULT_BASE_URL
        self.internal_base_url = DEFAULT_INTERNAL_BASE_URL
        self.health_url = None

        self.admin_username = "admin"
        self.admin_password = configuration.NONE_STRING
        self.app_secret = configuration.NONE_STRING

    def is_active(self):
        return self.port is not None


class BaseWebServer(object):

    def __init__(self, p_name, p_config, p_package_name, p_use_login_manager=False,
                 p_login_view=None, p_logged_out_endpoint=None):

        self._login_manager = None
        self._flask_stopper = None
        self._auth_view_handler = None
        self._server_started = False

        self._name = p_name
        self._config = p_config
        self._login_view = p_login_view

        # self._blueprint = flask.Blueprint("main", python_base_app.__name__)

        if p_package_name is None:
            raise configuration.ConfigurationException("HttpServer: p_package_name must not be None")

        self._logger = log_handling.get_logger(self.__class__.__name__)

        self._app = flask.Flask(p_package_name)
        self._flask_stopper = flask_helpers.FlaskStopper(p_app=self._app, p_logger=self._logger)

        self._app.config["APPLICATION_ROOT"] = self._config.base_url

        if p_use_login_manager:
            tools.check_config_value(p_config=self._config, p_config_attribute_name="admin_username")
            tools.check_config_value(p_config=self._config, p_config_attribute_name="admin_password")
            tools.check_config_value(p_config=self._config, p_config_attribute_name="app_secret")

            self._auth_view_handler = auth_view_handler.AuthViewHandler(
                p_admin_username=self._config.admin_username, p_admin_password=self._config.admin_password,
                p_app=self._app, p_logged_out_endpoint=p_logged_out_endpoint,
                p_url_prefix=self._config.base_url,
                p_login_view=p_login_view)

            self._app.config.update(SECRET_KEY=self._config.app_secret)

        self._server_exception = None

        # Install the actuator handler for the health check
        self._actuator_view_handler = actuator.ActuatorViewHandler(p_app=self._app, p_url_prefix=self._config.internal_base_url)

        logger = log_handling.get_logger("werkzeug")
        logger.setLevel(logging.WARNING)

        logger = log_handling.get_logger("sqlalchemy.engine")
        logger.setLevel(logging.WARNING)

    def destroy(self):
        self._actuator_view_handler.destroy()
        self._flask_stopper.destroy()

        if self._auth_view_handler is not None:
            self._auth_view_handler.destroy()

    def add_url_rule(self, p_rel_url, p_endpoint, p_view_method, p_blueprint=None,
                     p_methods=None,
                     p_login_required=False):

        if p_login_required:
            p_view_method = flask_login.login_required(p_view_method)

        blueprint = p_blueprint if p_blueprint is not None else self._blueprint

        blueprint.add_url_rule(join(self._config.base_url, p_rel_url),
                               p_endpoint,
                               p_view_method,
                               methods=p_methods)

    def get_url(self, p_rel_url='', p_internal=False, p_simple=False):

        return urllib.parse.urlunsplit(
            (
                self._config.scheme,
                "%s:%d" % (self._config.host, self._config.port),
                join(self._config.base_url, p_rel_url),
                None,
                None
            ))

    def start_server(self):

        tools.check_config_value(p_config=self._config, p_config_attribute_name="port")

        self._process = threading.Thread(target=self.run)
        self._process.start()
        time.sleep(1)

        if self._server_exception is not None:
            raise self._server_exception

        self._server_started = True

    def stop_server(self):

        if self._server_started and self._flask_stopper is not None:
            fmt = "Stopping web server %s..." % self._name
            self._logger.info(fmt)

            self._flask_stopper.stop(host=self._config.host, port=self._config.port)

            fmt = "Waiting for the server thread to terminate"
            self._logger.info(fmt)

            self._process.join()

            fmt = "HTTP server '%s' shut down successfully" % self._name
            self._logger.info(fmt)

    def run(self):

        tools.check_config_value(p_config=self._config, p_config_attribute_name="host")

        fmt = "Starting web server '{name}' on {address}:{port}"
        self._logger.info(fmt.format(name=self._name, address=self._config.host, port=self._config.port))

        # See https://stackoverflow.com/questions/14814201/can-i-serve-multiple-clients-using-just-flask-app-run-as-standalone
        try:
            self._app.run(port=self._config.port, host=self._config.host, threaded=True)

        except Exception as e:
            fmt = "Exception '%s' while starting web server %s" % (str(e), self._name)
            self._logger.error(fmt)

            self._server_exception = e
            raise e

    def get_authenication_info(self):

        return {
            "is_authenticated": flask_login.current_user is not None and flask_login.current_user.is_authenticated,
            "username": getattr(flask_login.current_user, "username",
                                "NO USER") if flask_login.current_user is not None else "NO USER"}



# -*- coding: utf-8 -*-

# Copyright (C) 2019  Marcus Rickert
#
# See https://github.com/marcus67/python_base_app
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import abc
import datetime
import locale

from python_base_app import configuration
from python_base_app import log_handling
from python_base_app import tools

REPLACE_PATTERN_AUDIO_TEXT_FILENAME = "infile"
REPLACE_PATTERN_BINARY = "binary"
REPLACE_PATTERN_AUDIO_TEXT = "text"
REPLACE_PATTERN_ENCODING = "encoding"

DEFAULT_MINIMUM_WAIT_BEFORE_REPEAT = 30  # seconds


class NotificationHandlerConfigModel(configuration.ConfigModel):

    def __init__(self, p_section_name):
        super().__init__(p_section_name=p_section_name)

        self.cache_audio_files = False
        self.mininum_waiting_time_before_repeat = DEFAULT_MINIMUM_WAIT_BEFORE_REPEAT  # seconds
        self.locale = None


class NotificationHandler(object):

    def __init__(self, p_config):

        self._config = p_config

        self._logger = log_handling.get_logger(self.__class__.__name__)

        self._recent_texts = {}

        if self._config.locale is None:
            self._config.locale = locale.getdefaultlocale()[0]

        self.init_engine()

    def init_engine(self):

        pass

    def stop_engine(self):

        pass

    def notify(self, p_text, p_locale=None):

        last_time_spoken = self._recent_texts.get(p_text)

        if p_locale is None:
            p_locale = self._config.locale

        if last_time_spoken is not None:
            if datetime.datetime.now() < last_time_spoken + datetime.timedelta(
                    seconds=self._config.mininum_waiting_time_before_repeat):
                fmt = "Text '{text}' has been output less than {waiting_time} seconds ago -> ignoring"
                self._logger.info(fmt.format(text=p_text, waiting_time=self._config.mininum_waiting_time_before_repeat))
                return None

        a_thread =  tools.start_simple_thread(method=self._notify, p_text=p_text, p_locale=p_locale)
        self._recent_texts[p_text] = datetime.datetime.now()

        return a_thread

    @abc.abstractmethod
    def _notify(self, p_text, p_locale=None):
        pass
